/*
 * GOCLOCK - 対局時計
 * main.c: メインソースファイル
 */

#include "common.h"

#include "display.h"
#include "gu.h"
#include "sys.h"

/*************************************************************************/
/************************ PSP module information *************************/
/*************************************************************************/

#define MODULE_FLAGS     0
#define MODULE_VERSION   0
#define MODULE_REVISION  1

PSP_MODULE_INFO("goclock", MODULE_FLAGS, MODULE_VERSION, MODULE_REVISION);
const PSP_MAIN_THREAD_PRIORITY(THREADPRI_MAIN);
const PSP_MAIN_THREAD_STACK_SIZE_KB(128);
const PSP_MAIN_THREAD_ATTR(PSP_THREAD_ATTR_USER);
const PSP_HEAP_SIZE_KB(-64);  // Leave 64k for thread stacks

/*************************************************************************/
/****************************** Local data *******************************/
/*************************************************************************/

/* 対局時間（分） */
static uint32_t start_time;

/* 残り時間（フレーム） */
static int32_t remaining[2];

/* リセット後、開始前かどうか */
static int is_reset;

/* 時計設定中かどうか */
static int setting;

/* 時計設定時の点滅タイマー */
static int setting_timer;

/* 時計が動いているかどうか */
static int running;

/* どちらの時計が動いているか（０＝左側） */
static int which_side;

/*----------------------------------*/

/* 文字テキスチャポインタとサイズ */
static uint8_t *font_texture;
static uint32_t font_texture_w, font_texture_h;

/* 音声チャンネル */
static int sound_channel;

/* 現在の音量 */
static volatile int sound_volume;

/* 音声データ */
static __attribute__((aligned(64))) int16_t sound_data[512];

/*-----------------------------------------------------------------------*/

/* Local routine declarations */

static void set_base_directory(const char *argv0);
static void read_font(void);
static void start_sound(void);

static void run_clock(void);
static void update_sound(void);
static void draw_frame(void);

static int frames_to_msec(int frames);
static void update_buttons(void);
static int check_button(int button, int new);
static void sound_thread(void);

/*************************************************************************/
/************************** Program entry point **************************/
/*************************************************************************/

/**
 * main:  Program entry point.  Performs initialization and then loops
 * indefinitely, running the emulator.
 *
 * [Parameters]
 *     argc: Command line argument count
 *     argv: Command line argument vector
 * [Return value]
 *     Does not return
 */
int main(int argc, char **argv)
{
    /* 電池消費を抑えるためにクロックを思いっきり下げる */
    scePowerSetClockFrequency(50, 50, 25);

    sys_setup_callbacks();
    display_init();
    set_base_directory(argv[0]);
    read_font();
    start_sound();
    is_reset = 1;
    setting = 0;
    running = 0;
    start_time = 0;
    remaining[0] = remaining[1] = 0;

    for (;;) {
        display_begin_frame();
        update_buttons();
        run_clock();
        update_sound();
        draw_frame();
        display_end_frame();
        scePowerTick(0);  // 自動画面オフを回避
    }
}

/*************************************************************************/
/****************************** 初期化関数 *******************************/
/*************************************************************************/

/**
 * set_base_directory:  現在ディレクトリをプログラムパス名から設定する。
 *
 * 【引　数】argv0: プログラムのパス名
 * 【戻り値】なし
 */
static void set_base_directory(const char *argv0)
{
    char buf[1000];
    const char *last_slash = strrchr(argv0, '/');
    const unsigned int path_len =
        last_slash ? last_slash - argv0 : strlen(argv0);
    if (path_len > sizeof(buf)-1) {
        DMSG("Path too long: %s", argv0);
        sceKernelExitGame();
    }
    memcpy(buf, argv0, path_len);
    buf[path_len] = 0;
    sceIoChdir(buf);
}

/*************************************************************************/

/**
 * read_font:  文字テキスチャをファイルから読み込む。
 *
 * 【引　数】なし
 * 【戻り値】なし
 */
static void read_font(void)
{
    font_texture = display_spare_vram();

    int fd = sceIoOpen("digits.bmp", PSP_O_RDONLY, 0);
    if (fd < 0) {
        return;
    }
    uint8_t buf[54];
    sceIoRead(fd, buf, 54);
    font_texture_w = buf[18] | buf[19]<<8;
    font_texture_h = buf[22] | buf[23]<<8;

    uint32_t *ptr =
        (uint32_t *)font_texture + (font_texture_w * (font_texture_h-1));
    int y;
    for (y = 0; y < font_texture_h; y++, ptr -= font_texture_w) {
        sceIoRead(fd, ptr, font_texture_w * 4);
    }

    sceIoClose(fd);
}

/*************************************************************************/

/**
 * start_sound:  音声データを生成し、再生を開始する。音量の調整で音を鳴らす。
 *
 * 【引　数】なし
 * 【戻り値】なし
 */
static void start_sound(void)
{
    int i;
    for (i = 0; i < lenof(sound_data); i++) {
        int pos = i*12 / lenof(sound_data);  // ~518Hz
        sound_data[i] = pos & 1 ? -0x1000 : 0x1000;
    }

    sound_channel = sceAudioChReserve(PSP_AUDIO_NEXT_CHANNEL,
                                      lenof(sound_data), 16 /*FORMAT_MONO*/);
    sound_volume = 0;
    sys_start_thread(
        "GoClockSoundThread", sound_thread, THREADPRI_SOUND,
        0x1000, 0, NULL
    );
}

/*************************************************************************/
/*************************** メインループ関数 ****************************/
/*************************************************************************/

/**
 * run_clock:  時計関連処理を行う。
 *
 * 【引　数】なし
 * 【戻り値】なし
 */
static void run_clock(void)
{
    if (running) {
        if (check_button(which_side ? PSP_CTRL_RTRIGGER : PSP_CTRL_LTRIGGER, 1)) {
            which_side ^= 1;
        } else if (check_button(PSP_CTRL_START, 1)) {
            running = 0;
        }
        remaining[which_side]--;
        if (remaining[which_side] <= 0) {
            running = 0;
        }
    } else {
        if (check_button(PSP_CTRL_SELECT, 0)) {
            setting = 1;
            if (check_button(PSP_CTRL_SELECT, 1)) {
                setting_timer = 0;
            }
            if (check_button(PSP_CTRL_UP, 1)) {
                start_time = (start_time + 10) % 600;
                setting_timer = 0;
            } else if (check_button(PSP_CTRL_DOWN, 1)) {
                start_time = (start_time + (600-10)) % 600;
                setting_timer = 0;
            } else if (check_button(PSP_CTRL_LEFT, 1)) {
                start_time = (start_time + (600-1)) % 600;
                setting_timer = 0;
            } else if (check_button(PSP_CTRL_RIGHT, 1)) {
                start_time = (start_time + 1) % 600;
                setting_timer = 0;
            }
            remaining[0] = remaining[1] =
                ((uint32_t)start_time * 60 * 60000) / 1001;
            is_reset = 1;
        } else {
            setting = 0;
            if (remaining[0] > 0 && remaining[1] > 0) {
                if (check_button(PSP_CTRL_START, 1) && !is_reset) {
                    running = 1;
                } else if (check_button(PSP_CTRL_LTRIGGER, 1)) {
                    is_reset = 0;
                    which_side = 1;
                    running = 1;
                } else if (check_button(PSP_CTRL_RTRIGGER, 1)) {
                    is_reset = 0;
                    which_side = 0;
                    running = 1;
                }
            }
        }
    }
}

/*************************************************************************/

/**
 * update_sound:  時計の状態に応じて、音声の状態を更新する。
 *
 * 【引　数】なし
 * 【戻り値】なし
 */
static void update_sound(void)
{
    int volume = 0;
    if (running) {
        const uint32_t msec = frames_to_msec(remaining[which_side]);
        if ((msec <= 60000 && msec > 59900)
         || (msec <= 30000 && msec > 29900)
         || (msec <= 20000 && msec > 19900)
         || (msec <= 10000 && msec >  9900)
         || (msec <=  9000 && msec >  8900)
         || (msec <=  8000 && msec >  7900)
         || (msec <=  7000 && msec >  6900)
         || (msec <=  6000 && msec >  5900)
         || (msec <=  5000 && msec >     0)
        ) {
            volume = PSP_AUDIO_VOLUME_MAX;
        }
    }
    sound_volume = volume;
}

/*************************************************************************/

/**
 * draw_frame:  画面を描画する。
 *
 * 【引　数】なし
 * 【戻り値】なし
 */
static void draw_frame(void)
{
    const int left_x0 = 52, right_x0 = 284, y0 = (272-64)/2;
    const int width = 144, height = 64;

    guDisable(GU_TEXTURE_2D);
    guDisable(GU_BLEND);

    /* 左右どちらの番かを示す */
    if (!is_reset) {
        const int x0 = which_side ? right_x0 : left_x0;
        const int x1 = x0 + width, y1 = y0 + height;
        struct {uint32_t color; uint16_t x, y, z, pad;} *vertices;
        vertices = guGetMemory(sizeof(*vertices) * 4);
        vertices[0].color = 0xFFFFFFFF;
        vertices[0].x = x0-35;
        vertices[0].y = y0-35;
        vertices[0].z = 0;
        vertices[1].color = 0xFFFFFFFF;
        vertices[1].x = x1+35;
        vertices[1].y = y1+35;
        vertices[1].z = 0;
        vertices[2].color = 0xFF000000;
        vertices[2].x = x0-20;
        vertices[2].y = y0-20;
        vertices[2].z = 0;
        vertices[3].color = 0xFF000000;
        vertices[3].x = x1+20;
        vertices[3].y = y1+20;
        vertices[3].z = 0;
        guDrawArray(GU_SPRITES,
                    GU_TRANSFORM_2D | GU_COLOR_8888 | GU_VERTEX_16BIT,
                    4, NULL, vertices);
        guCommit();
    }

    /* 時計設定中の点滅チェック */
    if (setting) {
        setting_timer = (setting_timer + 1) % 60;
        if (setting_timer >= 30) {
            return;  // 消えている
        }
    }

    /* 時計を描画する */
    guEnable(GU_TEXTURE_2D);
    guEnable(GU_BLEND);
    guBlendFunc(GU_ADD, GU_SRC_ALPHA, GU_ONE_MINUS_SRC_ALPHA, 0, 0);
    guTexFunc(GU_TFX_MODULATE, GU_TCC_RGBA);
    guTexMode(GU_PSM_8888, 0, 0, 0);
    guTexImage(0, 512, 512, font_texture_w, font_texture);
    int which;
    for (which = 0; which < 2; which++) {
        const int msec = frames_to_msec(remaining[which]);
        const int sec = (msec + 999) / 1000;  // Round up
        const int x0 = which ? right_x0 : left_x0;
        const int y1 = y0 + height;
        const int color = !is_reset && msec<=59000 ? 0xFF00FFFF : 0xFFFFFFFF;
        const int show_colon = !running || which != which_side || (msec % 1000 >= 500);
        int digits[4];
        digits[1] = show_colon ? 10 : -1;
        if (is_reset || sec >= 600) {
            const int hr = sec/3600;
            const int min = (sec/60) % 60;
            digits[0] = hr;
            digits[2] = min/10;
            digits[3] = min%10;
        } else {
            digits[0] = sec>=60 ? sec/60 : -1;
            digits[2] = sec>=10 ? (sec%60)/10 : -1;
            digits[3] = sec%10;
        }
        const int x[] = {x0, x0+40, x0+64, x0+104, x0+144};
        int i;
        for (i = 0; i < 4; i++) {
            if (digits[i] < 0) {
                continue;
            }
            struct {uint16_t u, v; uint32_t color; uint16_t x, y, z, pad;} *vertices;
            vertices = guGetMemory(sizeof(*vertices) * 2);
            vertices[0].u = digits[i] * 40;
            vertices[0].v = 0;
            vertices[0].color = color;
            vertices[0].x = x[i];
            vertices[0].y = y0;
            vertices[0].z = 0;
            vertices[1].u = digits[i] * 40 + (x[i+1] - x[i]);
            vertices[1].v = font_texture_h;
            vertices[1].color = color;
            vertices[1].x = x[i+1];
            vertices[1].y = y1;
            vertices[1].z = 0;
            guDrawArray(GU_SPRITES,
                        GU_TRANSFORM_2D | GU_TEXTURE_16BIT | GU_COLOR_8888 | GU_VERTEX_16BIT,
                        2, NULL, vertices);
        }
        guCommit();
    }
}

/*************************************************************************/
/***************************** その他の関数 ******************************/
/*************************************************************************/

/**
 * frames_to_msec:  フレーム数をミリ秒に変換する。
 *
 * 【引　数】frames: フレーム数
 * 【戻り値】floor(frames / (60/1.001) * 1000)
 */
static int frames_to_msec(int frames)
{
    return (unsigned int)frames * 1001 / 60;
}

/*************************************************************************/

static int cur_buttons;   // 現在の状態
static int last_buttons;  // 前フレームの状態

/*----------------------------------*/

/**
 * update_buttons:  ボタン押下状態を更新する。
 *
 * 【引　数】なし
 * 【戻り値】なし
 */
static void update_buttons(void)
{
    last_buttons = cur_buttons;
    SceCtrlData pad_data;
    sceCtrlPeekBufferPositive(&pad_data, 1);
    cur_buttons = pad_data.Buttons;
}

/*----------------------------------*/

/**
 * check_button:  指定されたボタンが押されているかどうかを返す。
 *
 * 【引　数】button: チェックするボタン（PSP_CTRL_*）
 * 　　　　　   new: ０以外＝新しく押されたかどうかを取得、０＝押下状態を取得
 * 【戻り値】０以外＝指定されたボタンが押されている、０＝押されていない
 */
static int check_button(int button, int new)
{
    if (new) {
        return ~last_buttons & cur_buttons & button;
    } else {
        return cur_buttons & button;
    }
}

/*************************************************************************/

/**
 * sound_thread:  音声再生スレッド。ひたすらに音声データをドライバに提出する。
 *
 * 【引　数】なし
 * 【戻り値】なし
 */
static void sound_thread(void)
{
    for (;;) {
        sceAudioOutputBlocking(sound_channel, sound_volume, sound_data);
    }
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
