/*  src/psp/sys.c: PSP system-related functions
    Copyright 2009 Andrew Church

    This file is part of Yabause.

    Yabause is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Yabause is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Yabause; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
*/

#include "common.h"

#include "sys.h"

/*************************************************************************/

/* Local routine declarations */

static int callback_thread(void);
static int exit_callback(int arg1, int arg2, void *common);
static int power_callback(int unknown, int power_info, void *common);

/*************************************************************************/
/************************** Interface functions **************************/
/*************************************************************************/

/**
 * psp_strerror:  Convert a PSP error code into a string, like strerror().
 *
 * [Parameters]
 *     code: Error code
 * [Return value]
 *     Corresponding error string
 * [Notes]
 *     The returned string is stored in a static buffer, so it will be
 *     destroyed on the next call to psp_strerror().
 */
const char *psp_strerror(const int32_t code)
{
    static char return_buf[100];

    /* Give descriptive messages for a few common codes */
    const char *message = NULL;
    switch (code) {
        case SCE_KERNEL_ERROR_NODEV: message = "No such device";      break;
        case SCE_KERNEL_ERROR_BADF:  message = "Bad file descriptor"; break;
        case SCE_KERNEL_ERROR_INVAL: message = "Invalid argument";    break;
        case SCE_KERNEL_ERROR_NOMEM: message = "Out of memory";       break;
    }

    if (message) {
        snprintf(return_buf, sizeof(return_buf), "%08X: %s", code, message);
    } else {
        snprintf(return_buf, sizeof(return_buf), "%08X", code);
    }
    return return_buf;
}

/*************************************************************************/

/**
 * sys_start_thread:  Start a new thread, returning the created thread
 * handle.
 *
 * [Parameters]
 *          name: Thread name
 *         entry: Thread entry address (function pointer)
 *      priority: Thread priority
 *     stacksize: Thread stack size
 *          args: Size of data to pass as thread argument
 *          argp: Pointer to thread argument data
 * [Return value]
 *     Nonnegative = thread handle, negative = error code
 */
int32_t sys_start_thread(const char *name, void *entry, int priority,
                         int stacksize, int args, void *argp)
{
    if (!name || !entry || priority < 0 || stacksize < 0) {
        DMSG("Invalid parameters: %p[%s] %p %d %d %d %p",
             name, name ? name : "", entry, priority, stacksize, args, argp);
        return SCE_KERNEL_ERROR_INVAL;
    }

    SceUID handle = sceKernelCreateThread(name, entry, priority, stacksize,
                                          0, NULL);
    if (handle < 0) {
        DMSG("Failed to create thread \"%s\": %s", name, psp_strerror(handle));
        return handle;
    }

    int32_t res = sceKernelStartThread(handle, args, argp);
    if (res < 0) {
        DMSG("Failed to start thread \"%s\": %s", name, psp_strerror(res));
        sceKernelDeleteThread(handle);
        return res;
    }

    return handle;
}

/*************************************************************************/

/**
 * sys_setup_callbacks:  Set up the system callbacks (HOME button and
 * power status change).
 *
 * [Parameters]
 *     None
 * [Return value]
 *     Nonzero on success, zero on error
 */
int sys_setup_callbacks(void)
{
    /* Start the callback monitoring thread */
    SceUID thid = sys_start_thread(
        "GoClockCallbackThread", callback_thread, THREADPRI_SYSTEM_CB,
        0x1000, 0, NULL
    );
    if (thid < 0) {
        DMSG("sys_start_thread(callback_thread) failed: %s",
             psp_strerror(thid));
        return 0;
    }

    return 1;
}

/*************************************************************************/
/************* System callback thread and callback routines **************/
/*************************************************************************/

/**
 * callback_thread:  Thread for monitoring HOME button and power status
 * notification callbacks.
 *
 * [Parameters]
 *     None
 * [Return value]
 *     Does not return
 */
static int callback_thread(void)
{
    SceUID cbid;
    cbid = sceKernelCreateCallback("GoClockExitCallback",
                                   exit_callback, NULL);
    if (cbid < 0) {
        DMSG("sceKernelCreateCallback(exit_callback) failed: %s",
             psp_strerror(cbid));
    } else {
        sceKernelRegisterExitCallback(cbid);
    }

    for (;;) {
        sceKernelSleepThreadCB();
    }

    return 0;  // Avoid compiler warning
}

/*************************************************************************/

/**
 * exit_callback:  HOME button callback.  Exits the program immediately.
 *
 * [Parameters]
 *     arg1, arg2, common: Unused
 * [Return value]
 *     Does not return
 */
static int exit_callback(int arg1, int arg2, void *common)
{
    sceKernelExitGame();
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
